@@ -13,18 +13,19 @@ use std::io::{BufReader, Read};
 use std::sync::Arc;
 use vrp_core::models::common::SingleDimLoad;
 use vrp_core::models::Problem;
+use vrp_core::solver::hyper::*;
 use vrp_core::solver::mutation::*;
 use vrp_core::solver::population::*;
 use vrp_core::solver::{Builder, Telemetry, TelemetryMode};
-use vrp_core::utils::{Environment, Parallelism};
+use vrp_core::utils::{Environment, Parallelism, Random};
 
 /// An algorithm configuration.
 #[derive(Clone, Deserialize, Debug)]
 pub struct Config {
     /// Specifies evolution configuration.
     pub evolution: Option<EvolutionConfig>,
-    /// Specifies mutation operator type.
-    pub mutation: Option<MutationType>,
+    /// Specifies hyper heuristic type.
+    pub hyper: Option<HyperType>,
     /// Specifies algorithm termination configuration.
     pub termination: Option<TerminationConfig>,
     /// Specifies environment configuration.
@@ -45,6 +46,14 @@ pub struct EvolutionConfig {
 #[serde(tag = "type")]
 #[serde(rename_all = "camelCase")]
 pub enum PopulationType {
+    /// A greedy population keeps track only of one best-known individual.
+    #[serde(rename(deserialize = "greedy"))]
+    #[serde(rename_all = "camelCase")]
+    Greedy {
+        /// Selection size. Default is number of cpus.
+        selection_size: Option<usize>,
+    },
+
     /// A basic population which sorts individuals based on their
     /// dominance order.
     #[serde(rename(deserialize = "elitism"))]
@@ -103,25 +112,39 @@ pub enum SelectionType {
     },
 }
 
+/// A hyper heuristic configuration.
+#[derive(Clone, Deserialize, Debug)]
+#[serde(tag = "type")]
+pub enum HyperType {
+    /// A hyper heuristic which selects one mutation from the list based on its probability.
+    #[serde(rename(deserialize = "static-selective"))]
+    StaticSelective {
+        /// A collection of inner mutation operators (metaheuristics).
+        mutations: Vec<MutationType>,
+    },
+}
+
 /// A mutation operator configuration.
 #[derive(Clone, Deserialize, Debug)]
 #[serde(tag = "type")]
 pub enum MutationType {
-    /// A metaheuristic which is composition of other metaheuristics with their
-    /// probability weights.
-    #[serde(rename(deserialize = "composite"))]
-    Composite {
-        /// Probability.
-        probability: f64,
-        /// A collection of inner metaheuristics.
-        inners: Vec<MutationType>,
+    /// A metaheuristic which splits problem into smaller and solves them independently.
+    #[serde(rename(deserialize = "decomposition"))]
+    #[serde(rename_all = "camelCase")]
+    Decomposition {
+        /// Max routes to be selected in decomposed solution.
+        routes: MinMaxConfig,
+        /// Amount of attempts to repeat refinement.
+        repeat: usize,
+        /// Probability of mutation.
+        probability: MutationProbabilityType,
     },
 
     /// A local search heuristic.
     #[serde(rename(deserialize = "local-search"))]
     LocalSearch {
-        /// Probability of the group.
-        probability: f64,
+        /// Probability of mutation.
+        probability: MutationProbabilityType,
         /// Amount of times one of operators is applied.
         times: MinMaxConfig,
         /// Local search operator.
@@ -132,7 +155,7 @@ pub enum MutationType {
     #[serde(rename(deserialize = "ruin-recreate"))]
     RuinRecreate {
         /// Probability.
-        probability: f64,
+        probability: MutationProbabilityType,
         /// Ruin methods.
         ruins: Vec<RuinGroupConfig>,
         /// Recreate methods.
@@ -140,7 +163,61 @@ pub enum MutationType {
     },
 }
 
-/// A ruin method configuration
+/// A mutation method probability type
+#[derive(Clone, Deserialize, Debug)]
+#[serde(untagged)]
+pub enum MutationProbabilityType {
+    /// A scalar probability based type.
+    Scalar {
+        /// Probability value of the mutation.
+        scalar: f64,
+    },
+
+    /// A context specific probability type.
+    Context {
+        /// Threshold parameters.
+        threshold: ContextThreshold,
+        /// Selection phase specific parameters.
+        phases: Vec<ContextPhase>,
+    },
+}
+
+/// A context condition for `MutationProbabilityType`.
+#[derive(Clone, Deserialize, Debug)]
+pub struct ContextThreshold {
+    /// Min amount of jobs in individual.
+    pub jobs: usize,
+    /// Min amount of routes in individual.
+    pub routes: usize,
+}
+
+/// A selection phase filter for `MutationProbabilityType`.
+#[derive(Clone, Deserialize, Debug)]
+#[serde(tag = "type")]
+pub enum ContextPhase {
+    /// Initial selection phase.
+    #[serde(rename(deserialize = "initial"))]
+    Initial {
+        /// A chance defined by probability.
+        chance: f64,
+    },
+
+    /// Exploration search phase.
+    #[serde(rename(deserialize = "exploration"))]
+    Exploration {
+        /// A chance defined by probability.
+        chance: f64,
+    },
+
+    /// Exploitation search phase.
+    #[serde(rename(deserialize = "exploitation"))]
+    Exploitation {
+        /// A chance defined by probability.
+        chance: f64,
+    },
+}
+
+/// A ruin method configuration.
 #[derive(Clone, Deserialize, Debug)]
 pub struct RuinGroupConfig {
     /// Ruin methods.
@@ -296,7 +373,7 @@ pub struct NameWeight {
 
 impl Default for Config {
     fn default() -> Self {
-        Self { evolution: None, mutation: None, termination: None, environment: None, telemetry: None }
+        Self { evolution: None, hyper: None, termination: None, environment: None, telemetry: None }
     }
 }
 
@@ -320,6 +397,9 @@ fn configure_from_evolution(
         if let Some(variation) = &config.population {
             let default_selection_size = get_default_selection_size(environment.as_ref());
             let population = match &variation {
+                PopulationType::Greedy { selection_size } => {
+                    Box::new(Greedy::new(problem, selection_size.unwrap_or(default_selection_size), None))
+                }
                 PopulationType::Elitism { max_size, selection_size } => Box::new(Elitism::new(
                     problem,
                     environment.random.clone(),
@@ -382,10 +462,22 @@ fn configure_from_evolution(
     Ok(builder)
 }
 
-fn configure_from_mutation(mut builder: Builder, mutation_config: &Option<MutationType>) -> Result<Builder, String> {
-    if let Some(config) = mutation_config {
-        let mutation = create_mutation(&builder.config.problem, config)?.0;
-        builder = builder.with_mutation(mutation)
+fn configure_from_hyper(
+    mut builder: Builder,
+    hyper_config: &Option<HyperType>,
+    environment: Arc<Environment>,
+) -> Result<Builder, String> {
+    if let Some(config) = hyper_config {
+        match config {
+            HyperType::StaticSelective { mutations } => {
+                let mutation_group = mutations
+                    .iter()
+                    .map(|mutation| create_mutation(&builder.config.problem, environment.random.clone(), mutation))
+                    .collect::<Result<Vec<_>, _>>()?;
+                let static_selective = vrp_core::solver::hyper::StaticSelective::new(mutation_group);
+                builder = builder.with_hyper(Box::new(static_selective));
+            }
+        }
     }
 
     Ok(builder)
@@ -404,45 +496,76 @@ fn configure_from_termination(
     Ok(builder)
 }
 
-fn create_recreate_method(method: &RecreateMethod) -> (Box<dyn Recreate + Send + Sync>, usize) {
+fn create_recreate_method(method: &RecreateMethod) -> (Arc<dyn Recreate + Send + Sync>, usize) {
     match method {
-        RecreateMethod::Cheapest { weight } => (Box::new(RecreateWithCheapest::default()), *weight),
-        RecreateMethod::Farthest { weight } => (Box::new(RecreateWithFarthest::default()), *weight),
-        RecreateMethod::SkipBest { weight, start, end } => (Box::new(RecreateWithSkipBest::new(*start, *end)), *weight),
-        RecreateMethod::Blinks { weight } => (Box::new(RecreateWithBlinks::<SingleDimLoad>::default()), *weight),
-        RecreateMethod::Gaps { weight, min } => (Box::new(RecreateWithGaps::new(*min)), *weight),
-        RecreateMethod::Nearest { weight } => (Box::new(RecreateWithNearestNeighbor::default()), *weight),
-        RecreateMethod::Regret { weight, start, end } => (Box::new(RecreateWithRegret::new(*start, *end)), *weight),
+        RecreateMethod::Cheapest { weight } => (Arc::new(RecreateWithCheapest::default()), *weight),
+        RecreateMethod::Farthest { weight } => (Arc::new(RecreateWithFarthest::default()), *weight),
+        RecreateMethod::SkipBest { weight, start, end } => (Arc::new(RecreateWithSkipBest::new(*start, *end)), *weight),
+        RecreateMethod::Blinks { weight } => (Arc::new(RecreateWithBlinks::<SingleDimLoad>::default()), *weight),
+        RecreateMethod::Gaps { weight, min } => (Arc::new(RecreateWithGaps::new(*min)), *weight),
+        RecreateMethod::Nearest { weight } => (Arc::new(RecreateWithNearestNeighbor::default()), *weight),
+        RecreateMethod::Regret { weight, start, end } => (Arc::new(RecreateWithRegret::new(*start, *end)), *weight),
         RecreateMethod::Perturbation { weight, probability, min, max } => {
-            (Box::new(RecreateWithPerturbation::new(*probability, *min, *max)), *weight)
+            (Arc::new(RecreateWithPerturbation::new(*probability, *min, *max)), *weight)
         }
     }
 }
 
 fn create_mutation(
     problem: &Arc<Problem>,
+    random: Arc<dyn Random + Send + Sync>,
     mutation: &MutationType,
-) -> Result<(Arc<dyn Mutation + Send + Sync>, f64), String> {
+) -> Result<(Arc<dyn Mutation + Send + Sync>, MutationProbability), String> {
     Ok(match mutation {
         MutationType::RuinRecreate { probability, ruins, recreates } => {
-            let ruin = Box::new(CompositeRuin::new(ruins.iter().map(|g| create_ruin_group(problem, g)).collect()));
+            let ruin = Arc::new(CompositeRuin::new(ruins.iter().map(|g| create_ruin_group(problem, g)).collect()));
             let recreate =
-                Box::new(CompositeRecreate::new(recreates.iter().map(|r| create_recreate_method(r)).collect()));
-            (Arc::new(RuinAndRecreate::new(recreate, ruin)), *probability)
+                Arc::new(CompositeRecreate::new(recreates.iter().map(|r| create_recreate_method(r)).collect()));
+            (Arc::new(RuinAndRecreate::new(ruin, recreate)), create_mutation_probability(probability, random.clone()))
         }
         MutationType::LocalSearch { probability, times, operators: inners } => {
             let operator = create_local_search(times, inners);
-            (Arc::new(LocalSearch::new(operator)), *probability)
+            (Arc::new(LocalSearch::new(operator)), create_mutation_probability(probability, random.clone()))
         }
-        MutationType::Composite { probability, inners } => {
-            let inners =
-                inners.iter().map(|mutation| create_mutation(problem, mutation)).collect::<Result<Vec<_>, _>>()?;
+        MutationType::Decomposition { routes, repeat, probability } => {
+            if *repeat < 1 {
+                return Err(format!("repeat must be greater than 1. Specified: {}", repeat));
+            }
+            if routes.min < 2 {
+                return Err(format!("min routes must be greater than 2. Specified: {}", routes.min));
+            }
 
-            (Arc::new(CompositeMutation::new(vec![(inners, 1)])), *probability)
+            let mutation = vrp_core::solver::hyper::StaticSelective::create_default_mutation(problem.clone());
+            (
+                Arc::new(DecomposeSearch::new(mutation, (routes.min, routes.max), *repeat)),
+                create_mutation_probability(probability, random.clone()),
+            )
         }
     })
 }
 
+fn create_mutation_probability(
+    probability: &MutationProbabilityType,
+    random: Arc<dyn Random + Send + Sync>,
+) -> MutationProbability {
+    match probability {
+        MutationProbabilityType::Scalar { scalar } => create_scalar_mutation_probability(*scalar, random),
+        MutationProbabilityType::Context { threshold, phases } => create_context_mutation_probability(
+            threshold.jobs,
+            threshold.routes,
+            phases
+                .iter()
+                .map(|phase| match phase {
+                    ContextPhase::Initial { chance } => (SelectionPhase::Initial, *chance),
+                    ContextPhase::Exploration { chance } => (SelectionPhase::Exploration, *chance),
+                    ContextPhase::Exploitation { chance } => (SelectionPhase::Exploitation, *chance),
+                })
+                .collect(),
+            random,
+        ),
+    }
+}
+
 fn create_ruin_group(problem: &Arc<Problem>, group: &RuinGroupConfig) -> RuinGroup {
     (group.methods.iter().map(|r| create_ruin_method(problem, r)).collect(), group.weight)
 }
@@ -471,23 +594,23 @@ fn create_ruin_method(problem: &Arc<Problem>, method: &RuinMethod) -> (Arc<dyn R
     }
 }
 
-fn create_local_search(times: &MinMaxConfig, inners: &[LocalOperatorType]) -> Box<dyn LocalOperator + Send + Sync> {
+fn create_local_search(times: &MinMaxConfig, inners: &[LocalOperatorType]) -> Arc<dyn LocalOperator + Send + Sync> {
     let operators = inners
         .iter()
-        .map::<(Box<dyn LocalOperator + Send + Sync>, usize), _>(|op| match op {
+        .map::<(Arc<dyn LocalOperator + Send + Sync>, usize), _>(|op| match op {
             LocalOperatorType::InterRouteBest { weight, noise } => {
-                (Box::new(ExchangeInterRouteBest::new(noise.probability, noise.min, noise.max)), *weight)
+                (Arc::new(ExchangeInterRouteBest::new(noise.probability, noise.min, noise.max)), *weight)
             }
             LocalOperatorType::InterRouteRandom { weight, noise } => {
-                (Box::new(ExchangeInterRouteRandom::new(noise.probability, noise.min, noise.max)), *weight)
+                (Arc::new(ExchangeInterRouteRandom::new(noise.probability, noise.min, noise.max)), *weight)
             }
             LocalOperatorType::IntraRouteRandom { weight, noise } => {
-                (Box::new(ExchangeIntraRouteRandom::new(noise.probability, noise.min, noise.max)), *weight)
+                (Arc::new(ExchangeIntraRouteRandom::new(noise.probability, noise.min, noise.max)), *weight)
             }
         })
         .collect::<Vec<_>>();
 
-    Box::new(CompositeLocalOperator::new(operators, times.min, times.max))
+    Arc::new(CompositeLocalOperator::new(operators, times.min, times.max))
 }
 
 fn configure_from_telemetry(builder: Builder, telemetry_config: &Option<TelemetryConfig>) -> Result<Builder, String> {
@@ -566,8 +689,8 @@ pub fn create_builder_from_config(problem: Arc<Problem>, config: &Config) -> Res
     let mut builder = Builder::new(problem.clone(), environment.clone());
 
     builder = configure_from_telemetry(builder, &config.telemetry)?;
-    builder = configure_from_evolution(builder, &config.evolution, problem, environment)?;
-    builder = configure_from_mutation(builder, &config.mutation)?;
+    builder = configure_from_evolution(builder, &config.evolution, problem, environment.clone())?;
+    builder = configure_from_hyper(builder, &config.hyper, environment)?;
     builder = configure_from_termination(builder, &config.termination)?;
 
     Ok(builder)
